import argparse
import collections
import copy
import csv
import json
import os
import re
import statistics
from pathlib import Path
from statistics import stdev
from typing import Dict, Tuple, Optional, List, Iterable

from randovania.game_description.default_database import game_description_for
from randovania.game_description.node import PickupNode, LogbookNode
from randovania.games.game import RandovaniaGame


def read_json(path: Path) -> dict:
    with path.open() as x:
        return json.load(x)


_KEY_MATCH = re.compile(r"Key (\d+)")
_PLAYER_MATCH = re.compile(r" for Player \d+")


def _filter_item_name(name: str) -> str:
    return _PLAYER_MATCH.sub("", _KEY_MATCH.sub("Key", name))


def accumulate_results(game_modifications: dict,
                       items: Dict[str, Dict[str, int]],
                       locations: Dict[str, Dict[str, int]],
                       item_hints: Dict[str, Dict[str, int]],
                       location_hints: Dict[str, Dict[str, int]],

                       index_to_location: Dict[int, Tuple[str, str]],
                       logbook_to_name: Dict[str, str],
                       ):
    for worlds, world_data in game_modifications["locations"].items():
        for area_name, item_name in world_data.items():
            item_name = _filter_item_name(item_name)
            items[item_name][area_name] += 1
            locations[area_name][item_name] += 1

    for logbook_asset, hint_data in game_modifications["hints"].items():
        if hint_data["hint_type"] != "location":
            continue

        logbook_asset = logbook_to_name[logbook_asset]

        if hint_data["target"] == -1:
            item_name = "Nothing"
        else:
            area_name, location_name = index_to_location[hint_data["target"]]
            item_name = game_modifications["locations"][area_name][location_name]

        item_name = _filter_item_name(item_name)
        location_hints[logbook_asset][item_name] += 1
        item_hints[item_name][logbook_asset] += 1


def calculate_pickup_count(items: Dict[str, Dict[str, int]]) -> Dict[str, int]:
    return {
        name: sum(data.values())
        for name, data in items.items()
    }


def sort_by_contents(data: dict) -> dict:
    return {
        item: {
            location: count
            for location, count in sorted(data[item].items(), key=lambda t: t[1], reverse=True)
        }
        for item in sorted(data.keys())
    }


def calculate_stddev(pickup_count: Dict[str, int], item_counts: Dict[str, float]) -> float:
    balanced_freq = {
        item: count / pickup_count[item]
        for item, count in item_counts.items()
    }
    return stdev(balanced_freq.values())


def first_key(d: dict):
    for key in d:
        return key


def get_items_order(all_items: Iterable[str], item_order: List[str]) -> Dict[str, int]:
    order = {
        item.split(" at ")[0]: i
        for i, item in enumerate(item_order)
    }
    for item in all_items:
        if item not in order:
            order[item] = len(item_order)

    return order


def create_report(seeds_dir: str, output_file: str, csv_dir: Optional[str]):
    def item_creator():
        return collections.defaultdict(int)

    items = collections.defaultdict(item_creator)
    locations = collections.defaultdict(item_creator)
    item_hints = collections.defaultdict(item_creator)
    location_hints = collections.defaultdict(item_creator)
    item_order = collections.defaultdict(list)

    game_description = game_description_for(RandovaniaGame.PRIME2)
    world_list = game_description.world_list
    index_to_location = {
        node.pickup_index.index: (world_list.world_name_from_node(node, distinguish_dark_aether=True),
                                  world_list.node_name(node))
        for node in game_description.world_list.all_nodes
        if isinstance(node, PickupNode)
    }

    logbook_to_name = {
        str(node.string_asset_id): game_description.world_list.node_name(node)
        for node in game_description.world_list.all_nodes
        if isinstance(node, LogbookNode)
    }

    seed_count = 0
    pickup_count = None
    for seed in Path(seeds_dir).glob("**/*.json"):
        seed_data = read_json(seed)
        for game_modifications in seed_data["game_modifications"]:
            accumulate_results(game_modifications,
                               items, locations,
                               item_hints, location_hints,
                               index_to_location, logbook_to_name)
        if seed_count == 0:
            pickup_count = calculate_pickup_count(items)

        for item, order in get_items_order(list(items.keys()), seed_data["item_order"]).items():
            item_order[item].append(order)

        seed_count += 1

    if pickup_count is None:
        raise Exception("No seeds found")

    stddev_by_location = {
        location: calculate_stddev(pickup_count, locations[location])
        for location in locations.keys()
    }

    final_results = {
        "seed_count": seed_count,
        "stddev_by_location": {
            location: stddev
            for location, stddev in sorted(stddev_by_location.items(), key=lambda t: t[1], reverse=True)
        },
        "items": sort_by_contents(items),
        "locations": sort_by_contents(locations),
        "item_hints": sort_by_contents(item_hints),
        "location_hints": sort_by_contents(location_hints),
        "item_order": {
            "average": {
                name: statistics.mean(orders)
                for name, orders in item_order.items()
            },
            "median": {
                name: int(statistics.median(orders))
                for name, orders in item_order.items()
            },
            "stdev": {
                name: statistics.stdev(orders)
                for name, orders in item_order.items()
            },
        }
    }

    if csv_dir is not None:
        os.makedirs(csv_dir, exist_ok=True)
        for field in "items", "locations", "item_hints", "location_hints":
            data = final_results[field]

            possible_columns = set()
            for potential_values in data.values():
                possible_columns |= set(potential_values.keys())

            possible_columns = list(sorted(possible_columns))
            possible_columns.insert(0, "row_name")

            with open(os.path.join(csv_dir, field + ".csv"), "w", newline='') as csv_file:
                writer = csv.DictWriter(csv_file, fieldnames=possible_columns)
                writer.writeheader()

                for column, row_data in data.items():
                    row_data = copy.copy(row_data)
                    row_data["row_name"] = column
                    writer.writerow(row_data)

    with open(output_file, "w") as output:
        json.dump(final_results, output, indent=4, separators=(',', ': '))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--csv-dir")
    parser.add_argument("seeds_dir")
    parser.add_argument("output_file")
    args = parser.parse_args()
    create_report(args.seeds_dir, args.output_file,
                  args.csv_dir)


if __name__ == "__main__":
    main()



from enum import Enum


class GameSessionState(Enum):
    """State of a game session"""
    SETUP = "setup"
    IN_PROGRESS = "in-progress"
    FINISHED = "finished"

    @property
    def user_friendly_name(self) -> str:
        return self.value.title()



import pytest

from randovania.layout.layout_description import LayoutDescription
from randovania.resolver import resolver, debug


@pytest.mark.skip_resolver_tests
def test_resolver_with_log_file(test_files_dir):
    # Setup
    debug.set_level(0)

    description = LayoutDescription.from_file(test_files_dir.joinpath("log_files", "seed_a.rdvgame"))
    configuration = description.permalink.presets[0].configuration
    patches = description.all_patches[0]

    # Run
    final_state_by_resolve = resolver.resolve(configuration=configuration,
                                              patches=patches)

    # Assert
    assert final_state_by_resolve is not None



from dataclasses import dataclass
from typing import Optional, Tuple

from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_index import PickupIndex


@dataclass(frozen=True)
class MajorItem:
    name: str
    item_category: ItemCategory
    broad_category: ItemCategory
    model_index: Optional[int]
    progression: Tuple[int, ...]
    ammo_index: Tuple[int, ...] = tuple()
    converts_indices: Tuple[int, ...] = tuple()
    required: bool = False
    original_index: Optional[PickupIndex] = None
    probability_offset: int = 0
    probability_multiplier: float = 1
    warning: Optional[str] = None

    @classmethod
    def from_json(cls, name: str, value: dict) -> "MajorItem":
        return cls(
            name=name,
            item_category=ItemCategory(value["item_category"]),
            broad_category=ItemCategory(value["broad_category"]),
            model_index=value["model_index"],
            progression=tuple(value["progression"]),
            ammo_index=tuple(value.get("ammo", [])),
            converts_indices=tuple(value.get("converts_indices", [])),
            required=value.get("required", False),
            original_index=PickupIndex(value["original_index"]) if "original_index" in value else None,
            probability_offset=value["probability_offset"],
            probability_multiplier=value["probability_multiplier"],
            warning=value.get("warning"),
        )

    @property
    def as_json(self) -> dict:
        result = {
            "item_category": self.item_category.value,
            "broad_category": self.broad_category.value,
            "model_index": self.model_index,
            "progression": list(self.progression),
            "ammo": list(self.ammo_index),
            "converts_indices": list(self.converts_indices),
            "required": self.required,
            "probability_offset": self.probability_offset,
            "probability_multiplier": self.probability_multiplier,
        }
        if self.original_index is not None:
            result["original_index"] = self.original_index.index
        if self.warning is not None:
            result["warning"] = self.warning
        return result



import dataclasses as _dataclasses
import struct as _struct
from typing import Optional, Callable, Tuple


def _pack(*args):
    return _struct.pack(*args)


@_dataclasses.dataclass(frozen=True)
class Register:
    number: int


class GeneralRegister(Register):
    pass


class FloatRegister(Register):
    pass


class BaseInstruction:
    label: Optional[str]

    def __init__(self):
        self.label = None

    def with_label(self, label: str) -> "BaseInstruction":
        self.label = label
        return self

    def bytes_for(self, address: int):
        raise NotImplementedError()

    def __eq__(self, other):
        raise NotImplementedError()

    @property
    def byte_count(self):
        raise NotImplementedError()


class Instruction(BaseInstruction):
    value: int

    def __init__(self, value: int):
        super().__init__()
        self.value = value

    def bytes_for(self, address: int):
        return iter(_pack(">I", self.value))

    def __eq__(self, other):
        return isinstance(other, Instruction) and self.value == other.value

    @property
    def byte_count(self):
        return 4

    @classmethod
    def compose(cls, data: Tuple[Tuple[int, int, bool], ...]):
        value = 0
        bits_left = 32
        for item, bit_size, signed in data:
            bits_left -= bit_size
            if signed:
                assert -(1 << (bit_size - 1)) <= item < (1 << (bit_size - 1))
                if item < 0:
                    item += (1 << bit_size)
            else:
                assert 0 <= item < (1 << bit_size)
            value += item << bits_left

        assert bits_left == 0
        return cls(value)


class AddressDependantInstruction(BaseInstruction):
    def __init__(self, factory: Callable[[int], Tuple[Tuple[int, int, bool], ...]]):
        super().__init__()
        self.factory = factory

    def bytes_for(self, address: int):
        yield from Instruction.compose(self.factory(address)).bytes_for(address)

    def __eq__(self, other):
        return isinstance(other, AddressDependantInstruction) and self.factory == other.factory

    @property
    def byte_count(self):
        return 4


r0 = GeneralRegister(0)
r1 = GeneralRegister(1)
r2 = GeneralRegister(2)
r3 = GeneralRegister(3)
r4 = GeneralRegister(4)
r5 = GeneralRegister(5)
r6 = GeneralRegister(6)
r7 = GeneralRegister(7)
r8 = GeneralRegister(8)
r9 = GeneralRegister(9)
r10 = GeneralRegister(10)
r11 = GeneralRegister(11)
r12 = GeneralRegister(12)
r25 = GeneralRegister(25)
r26 = GeneralRegister(26)
r27 = GeneralRegister(27)
r28 = GeneralRegister(28)
r29 = GeneralRegister(29)
r30 = GeneralRegister(30)
r31 = GeneralRegister(31)

f0 = FloatRegister(0)
f1 = FloatRegister(1)
f2 = FloatRegister(2)
f3 = FloatRegister(3)

# Special Registers
LR = 8
CTR = 9


def lmw(start_register: GeneralRegister, offset: int, input_register: GeneralRegister):
    return Instruction.compose(((46, 6, False),
                                (start_register.number, 5, False),
                                (input_register.number, 5, False),
                                (offset, 16, True)))


def lwz(output_register: GeneralRegister, offset: int, input_register: GeneralRegister):
    """
    *(output_register + offset) = input_register
    """
    return Instruction.compose(((32, 6, False),
                                (output_register.number, 5, False),
                                (input_register.number, 5, False),
                                (offset, 16, True)))


def lbz(output_register: GeneralRegister, offset: int, input_register: GeneralRegister):
    """
    *(output_register + offset) = input_register
    """
    return Instruction.compose(((34, 6, False),
                                (output_register.number, 5, False),
                                (input_register.number, 5, False),
                                (offset, 16, True)))


def or_(output_register: GeneralRegister, input_register_a: GeneralRegister, input_register_b: GeneralRegister,
        record_bit: bool = False):
    """
    output_register = input_register_a | input_register_b
    """
    # See https://www.ibm.com/support/knowledgecenter/en/ssw_aix_72/assembler/idalangref_or_instruction.html
    return Instruction.compose(((31, 6, False),
                                (input_register_a.number, 5, False),
                                (output_register.number, 5, False),
                                (input_register_b.number, 5, False),
                                (444, 10, False),
                                (int(record_bit), 1, False),
                                ))


def ori(output_register: GeneralRegister, input_register: GeneralRegister, constant: int):
    """
    output_register = input_register | constant
    """
    return Instruction.compose(((24, 6, False),
                                (output_register.number, 5, False),
                                (input_register.number, 5, False),
                                (constant, 16, False)))


def nop():
    return ori(r0, r0, 0x0)


def li(register: GeneralRegister, literal: int):
    """
    register = literal
    """
    return addi(register, r0, literal)


def lis(register: GeneralRegister, literal: int):
    """
    register = literal
    """
    return addis(register, r0, literal)


def lfs(output_register: FloatRegister, offset: int, input_register: GeneralRegister):
    """
    output_register = (float) *(input_register + offset)

    output_register is a float register.
    """
    return Instruction.compose(((48, 6, False),
                                (output_register.number, 5, False),
                                (input_register.number, 5, False),
                                (offset, 16, True)))


def cmpwi(input_register: GeneralRegister, literal: int):
    return Instruction.compose(((11, 6, False),
                                (0, 3, False),
                                (0, 1, False),
                                (0, 1, False),
                                (input_register.number, 5, False),
                                (literal, 16, True)))


def _jump_to_relative_address(address_or_symbol: int, link: bool):
    def with_inc_address(instruction_address: int):
        jump_offset = (address_or_symbol - instruction_address) // 4
        return (((18, 6, False),
                 (jump_offset, 24, True),
                 (0, 1, False),
                 (int(link), 1, False)))

    return AddressDependantInstruction(with_inc_address)


def b(address_or_symbol: int):
    """
    jumps to the given address, not setting the link register
    """
    return _jump_to_relative_address(address_or_symbol, False)


def bl(address_or_symbol: int):
    """
    jumps to the given address, setting the link register
    """
    return _jump_to_relative_address(address_or_symbol, True)


def _conditional_branch(bo: int, bi: int, address_or_symbol: int, relative: bool = False):
    # https://www.ibm.com/support/knowledgecenter/ssw_aix_72/assembler/idalangref_ext_br_mnem_bofield.html#idalangref_ext_br_mnem_bofield__row-d2e17648

    def with_inc_address(instruction_address: int):
        jump_offset = (address_or_symbol - instruction_address) // 4
        return (((16, 6, False),
                 (bo, 5, False),
                 (bi, 5, False),
                 (jump_offset, 14, True),
                 (0, 1, False),
                 (0, 1, False)))

    if relative:
        return Instruction.compose(with_inc_address(0))
    else:
        return AddressDependantInstruction(with_inc_address)


def beq(address_or_symbol: int, relative: bool = False):
    """
    jumps to the given address, if last comparison was a successful equality
    """
    bo = 12  # Branch if condition true (BO=12)
    bi = 2  # condition: equals
    return _conditional_branch(bo, bi, address_or_symbol, relative=relative)


def bne(address_or_symbol: int, relative: bool = False):
    """
    jumps to the given address, if last comparison was a successful equality
    """
    bo = 4  # Branch if condition true (BO=4)
    bi = 2  # condition: equals
    return _conditional_branch(bo, bi, address_or_symbol, relative=relative)


def bclr(bo, bi, bh):
    # https://www.ibm.com/support/knowledgecenter/ssw_aix_72/assembler/idalangref_branch_conditional_link_register.html
    lk = 0
    return Instruction.compose(((19, 6, False),
                                (bo, 5, False),
                                (bi, 5, False),
                                (0, 3, False),
                                (bh, 2, False),
                                (16, 10, False),
                                (lk, 1, False)))


def blr():
    """Branches to the address set by the link register. Does not set the link register."""
    return bclr(20, 0, 0)


def bcctrl(bo, bi, bh):
    """Branch conditionally. Sets the link register."""
    lk = 1
    return Instruction.compose(((19, 6, False),
                                (bo, 5, False),
                                (bi, 5, False),
                                (0, 3, False),
                                (bh, 2, False),
                                (528, 10, False),
                                (lk, 1, False)))


def bctrl():
    """Branches always. Sets the link register."""
    return bcctrl(20, 0, 0)


def _store(input_register: Register, offset: int, output_register: GeneralRegister, op_code: int):
    return Instruction.compose(((op_code, 6, False),
                                (input_register.number, 5, False),
                                (output_register.number, 5, False),
                                (offset, 16, True)))


def stb(input_register: GeneralRegister, offset: int, output_register: GeneralRegister):
    """
    *(output_register +offset) = input_register
    """
    return _store(input_register, offset, output_register, 38)


def stw(input_register: GeneralRegister, offset: int, output_register: GeneralRegister):
    """
    *(output_register +offset) = input_register
    """
    return _store(input_register, offset, output_register, 36)


def stfs(input_register: FloatRegister, offset: int, output_register: GeneralRegister):
    """
    *(float*)(output_register +offset) = input_register
    """
    return _store(input_register, offset, output_register, 52)


def stwu(input_register: GeneralRegister, offset: int, output_register: GeneralRegister):
    return _store(input_register, offset, output_register, 37)


def stmw(start_register: GeneralRegister, offset: int, output_register: GeneralRegister):
    return _store(start_register, offset, output_register, 47)


def sync():
    return Instruction(0x7c0004ac)


def icbi(ra: int, rb: int):
    value = 0x7C0007AC + (ra << 16) + (rb << 11)
    return Instruction(value)


def dcbi(ra: int, rb: int):
    return Instruction.compose(((31, 6, False),
                                (0, 5, False),
                                (ra, 5, False),
                                (rb, 5, False),
                                (470, 10, False),
                                (0, 1, False)))


def isync():
    return Instruction(0x4C00012C)


def addi(output_register: GeneralRegister, input_register: GeneralRegister, literal: int):
    """
    output_register = input_register + literal
    """
    return Instruction.compose(((14, 6, False),
                                (output_register.number, 5, False),
                                (input_register.number, 5, False),
                                (literal, 16, True)))


def addis(output_register: GeneralRegister, input_register: GeneralRegister, literal: int):
    """
    output_register = (input_register + literal) << 16
    """
    return Instruction.compose(((15, 6, False),
                                (output_register.number, 5, False),
                                (input_register.number, 5, False),
                                (literal, 16, False)))


def _special_register_op(input_register: Register, special_register: int, magic_value: int, op_code: int):
    special_register_top = special_register >> 5
    special_register_bot = special_register & 0b11111

    return Instruction.compose(((op_code, 6, False),
                                (input_register.number, 5, False),
                                (special_register_bot, 5, False),
                                (special_register_top, 5, False),
                                (magic_value, 10, False),
                                (0, 1, False),
                                ))


def mtspr(special_register, input_register: GeneralRegister):
    return _special_register_op(input_register, special_register, 467, 31)


def mtctr(rs: GeneralRegister):
    return mtspr(CTR, rs)


def mfspr(output_register: GeneralRegister, special_register):
    """
    Move from Special-Purpose Register
    https://www.ibm.com/support/knowledgecenter/ssw_aix_72/assembler/idalangref_mfspr_spr_instrs.html
    """
    return _special_register_op(output_register, special_register, 339, 31)



from PySide2 import QtCore

from randovania.gui.generated.preset_echoes_goal_ui import Ui_PresetEchoesGoal
from randovania.gui.lib.common_qt_lib import set_combo_with_value
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.echoes_configuration import LayoutSkyTempleKeyMode
from randovania.layout.preset import Preset


class PresetEchoesGoal(PresetTab, Ui_PresetEchoesGoal):
    _editor: PresetEditor

    def __init__(self, editor: PresetEditor):
        super().__init__()
        self.setupUi(self)
        self._editor = editor

        self.goal_layout.setAlignment(QtCore.Qt.AlignTop)
        self.skytemple_combo.setItemData(0, LayoutSkyTempleKeyMode.ALL_BOSSES)
        self.skytemple_combo.setItemData(1, LayoutSkyTempleKeyMode.ALL_GUARDIANS)
        self.skytemple_combo.setItemData(2, int)

        self.skytemple_combo.options_field_name = "layout_configuration_sky_temple_keys"
        self.skytemple_combo.currentIndexChanged.connect(self._on_sky_temple_key_combo_changed)
        self.skytemple_slider.valueChanged.connect(self._on_sky_temple_key_combo_slider_changed)

    @property
    def uses_patches_tab(self) -> bool:
        return False

    def _on_sky_temple_key_combo_changed(self):
        combo_enum = self.skytemple_combo.currentData()
        with self._editor:
            if combo_enum is int:
                self.skytemple_slider.setEnabled(True)
                self._editor.layout_configuration_sky_temple_keys = LayoutSkyTempleKeyMode(
                    self.skytemple_slider.value())
            else:
                self.skytemple_slider.setEnabled(False)
                self._editor.layout_configuration_sky_temple_keys = combo_enum

    def _on_sky_temple_key_combo_slider_changed(self):
        self.skytemple_slider_label.setText(str(self.skytemple_slider.value()))
        self._on_sky_temple_key_combo_changed()

    def on_preset_changed(self, preset: Preset):
        keys = preset.configuration.sky_temple_keys
        if isinstance(keys.value, int):
            self.skytemple_slider.setValue(keys.value)
            data = int
        else:
            data = keys
        set_combo_with_value(self.skytemple_combo, data)



import argparse
from unittest.mock import MagicMock

import pytest

from randovania.cli import server


def test_server_command_logic(mocker):
    # Setup
    mock_create_app: MagicMock = mocker.patch("randovania.server.app.create_app")

    # Run
    server.flask_command_logic(None)

    # Assert
    mock_create_app.assert_called_once_with()
    mock_create_app.return_value.sio.sio.run(mock_create_app.return_value, host="0.0.0.0")


def test_create_subparsers():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    server.create_subparsers(subparsers)

    with pytest.raises(SystemExit):
        args = parser.parse_args(["multiworld"])
        args.func(args)



from unittest.mock import MagicMock, PropertyMock

from randovania.game_description.node import EventNode
from randovania.resolver.resolver_reach import ResolverReach


def test_possible_actions_empty():
    state = MagicMock()

    reach = ResolverReach({}, {}, frozenset(), MagicMock())
    options = list(reach.possible_actions(state))

    assert options == []


def test_possible_actions_no_resources():
    state = MagicMock()
    node_a = MagicMock()
    node_b = MagicMock()
    node_b.can_collect.return_value = False

    type(node_a).is_resource_node = prop_a = PropertyMock(return_value=False)
    type(node_b).is_resource_node = prop_b = PropertyMock(return_value=True)

    # Run
    reach = ResolverReach({node_a: 1, node_b: 1}, {}, frozenset(), MagicMock())
    options = list(action for action, damage in reach.possible_actions(state))

    # Assert
    assert options == []
    prop_a.assert_called_once_with()
    prop_b.assert_called_once_with()
    node_b.can_collect.assert_called_once_with(state.patches, state.resources, reach._logic.game.world_list.all_nodes)


def test_possible_actions_with_event():
    logic = MagicMock()
    state = MagicMock()

    event = MagicMock(spec=EventNode)
    type(event).is_resource_node = prop = PropertyMock(return_value=True)
    event.can_collect.return_value = True

    # Run
    reach = ResolverReach({event: 1}, {}, frozenset(), logic)
    options = list(action for action, damage in reach.possible_actions(state))

    # Assert
    assert options == [event]
    prop.assert_called_once_with()
    event.can_collect.assert_called_once_with(state.patches, state.resources, reach._logic.game.world_list.all_nodes)
    logic.get_additional_requirements.assert_called_once_with(event)
    logic.get_additional_requirements.return_value.satisfied.assert_called_once_with(state.resources, 1)



import pytest

from randovania.dol_patching import assembler
from randovania.dol_patching.assembler import custom_ppc, ppc


@pytest.mark.parametrize("value", [-1, 0x100000000])
def test_load_unsigned_32bit_invalid(value):
    with pytest.raises(AssertionError):
        custom_ppc.load_unsigned_32bit(ppc.r1, value)


@pytest.mark.parametrize(["value", "expected"], [
    (0x00000000, b"\x3c\x20\x00\x00\x60\x21\x00\x00"),
    (0xFFFFFFFF, b"\x3c\x20\xFF\xFF\x60\x21\xFF\xFF"),
])
def test_load_unsigned_32bit_valid(value, expected):
    codes = list(custom_ppc.load_unsigned_32bit(ppc.r1, value).bytes_for(0))
    assert bytes(codes) == expected


def test_composite_bl_single():
    inc = ppc.bl(0x80085760)
    comp = custom_ppc.CompositeInstruction((inc,))

    assert list(comp.bytes_for(0x80085760 + 0x9C)) == [0x4b, 0xff, 0xff, 0x65]


def test_composite_bl_double():
    target_address = 0x80085760
    start_address = 0x80085760 + 0x9C

    inc = ppc.bl(target_address)
    comp = custom_ppc.CompositeInstruction((inc, inc))

    composite_bytes = list(comp.bytes_for(start_address))
    assembled_bytes = list(assembler.assemble_instructions(start_address, (inc, inc)))

    assert composite_bytes == [0x4b, 0xff, 0xff, 0x65, 0x4b, 0xff, 0xff, 0x61]
    assert composite_bytes == assembled_bytes



import copy
import json
from pathlib import Path
from typing import Optional, Dict

import aiofiles
import slugify

from randovania.games.game import RandovaniaGame
from randovania.layout.preset import Preset

CURRENT_PRESET_VERSION = 8


class InvalidPreset(Exception):
    def __init__(self, original_exception: Exception):
        self.original_exception = original_exception


def _migrate_v1(preset: dict) -> dict:
    layout_configuration = preset["layout_configuration"]
    layout_configuration["beam_configuration"] = {
        "power": {
            "item_index": 0,
            "ammo_a": -1,
            "ammo_b": -1,
            "uncharged_cost": 0,
            "charged_cost": 0,
            "combo_missile_cost": 5,
            "combo_ammo_cost": 0
        },
        "dark": {
            "item_index": 1,
            "ammo_a": 45,
            "ammo_b": -1,
            "uncharged_cost": 1,
            "charged_cost": 5,
            "combo_missile_cost": 5,
            "combo_ammo_cost": 30
        },
        "light": {
            "item_index": 2,
            "ammo_a": 46,
            "ammo_b": -1,
            "uncharged_cost": 1,
            "charged_cost": 5,
            "combo_missile_cost": 5,
            "combo_ammo_cost": 30
        },
        "annihilator": {
            "item_index": 3,
            "ammo_a": 46,
            "ammo_b": 45,
            "uncharged_cost": 1,
            "charged_cost": 5,
            "combo_missile_cost": 5,
            "combo_ammo_cost": 30
        }
    }
    layout_configuration["skip_final_bosses"] = False
    layout_configuration["energy_per_tank"] = 100.0
    return preset


def _migrate_v2(preset: dict) -> dict:
    level_renaming = {
        "trivial": "beginner",
        "easy": "intermediate",
        "normal": "advanced",
        "hard": "expert",
        "minimal-restrictions": "minimal-logic",
    }
    trick_level = preset["layout_configuration"]["trick_level"]
    trick_level["global_level"] = level_renaming.get(trick_level["global_level"], trick_level["global_level"])
    for specific, value in trick_level["specific_levels"].items():
        trick_level["specific_levels"][specific] = level_renaming.get(value, value)

    return preset


def _migrate_v3(preset: dict) -> dict:
    preset["layout_configuration"]["safe_zone"] = {
        "fully_heal": True,
        "prevents_dark_aether": True,
        "heal_per_second": 1.0,
    }
    return preset


def _migrate_v4(preset: dict) -> dict:
    trick_name_mapping = {
        0: "Dash",
        1: "BombJump",
        2: "SlopeJump",
        3: "Movement",
        4: "BSJ",
        5: "RollJump",
        6: "UnderwaterDash",
        7: "AirUnderwater",
        8: "OoB",
        10: "SAnoSJ",
        11: "WallBoost",
        12: "EnemyHop",
        13: "Combat",
        15: "InstantMorph",
        26: "InvisibleObjects",
        27: "StandableTerrain",
        28: "TerminalFall",
        29: "BoostJump",
        30: "EDash",
        31: "BomblessSlot",
        32: "ScanPost",
        33: "ScrewAttackTunnels",
        34: "Knowledge",
        35: "SeekerlessLocks",
    }

    preset["layout_configuration"]["game"] = "prime2"

    trick_level = preset["layout_configuration"]["trick_level"]
    global_level = trick_level.pop("global_level")

    trick_level["minimal_logic"] = global_level == "minimal-logic"
    specific_levels = {
        trick_name_mapping[int(index)]: level
        for index, level in trick_level["specific_levels"].items()
    }
    if global_level == "minimal-logic":
        specific_levels = {}
    elif global_level == "beginner":
        for trick in ["BombJump", "Knowledge", "Movement", "ScanPost", "SAnoSJ"]:
            specific_levels[trick] = "beginner"
    else:
        for trick in trick_name_mapping.values():
            if trick not in specific_levels:
                specific_levels[trick] = global_level

    trick_level["specific_levels"] = specific_levels

    return preset


def _migrate_v5(preset: dict) -> dict:
    excluded_item = {
        "include_copy_in_original_location": False,
        "num_shuffled_pickups": 0,
        "num_included_in_starting_items": 0,
        "included_ammo": [],
        "allowed_as_random_starting_item": True
    }
    included_item = {
        **excluded_item,
        "num_included_in_starting_items": 1
    }
    shuffled_item = {
        **excluded_item,
        "num_shuffled_pickups": 1,
    }

    default_items_state = {
        "Progressive Suit": {**excluded_item, "num_shuffled_pickups": 2},
        "Dark Beam": {**shuffled_item, "included_ammo": [50]},
        "Light Beam": {**shuffled_item, "included_ammo": [50]},
        "Annihilator Beam": {**shuffled_item, "included_ammo": [0, 0]},
        "Power Bomb": {**shuffled_item, "included_ammo": [2]},
        "Progressive Grapple": {**excluded_item, "num_shuffled_pickups": 2},
        "Missile Launcher": {**shuffled_item, "included_ammo": [5]},
        "Seeker Launcher": {**shuffled_item, "included_ammo": [5]},
        "Energy Tank": {**excluded_item, "num_shuffled_pickups": 14},
    }

    for item in ["Combat Visor", "Scan Visor", "Varia Suit", "Power Beam", "Charge Beam", "Morph Ball"]:
        default_items_state[item] = included_item
    for item in ["Dark Visor", "Echo Visor", "Morph Ball Bomb", "Boost Ball", "Spider Ball", "Space Jump Boots",
                 "Gravity Boost", "Super Missile", "Sunburst", "Darkburst", "Sonic Boom", "Violet Translator",
                 "Amber Translator", "Emerald Translator", "Cobalt Translator"]:
        default_items_state[item] = shuffled_item

    major_items = preset["layout_configuration"]["major_items_configuration"]["items_state"]
    for item in default_items_state.keys():
        if item not in major_items:
            major_items[item] = default_items_state[item]

    preset["layout_configuration"]["major_items_configuration"].pop("progressive_suit")
    preset["layout_configuration"]["major_items_configuration"].pop("progressive_grapple")
    preset["layout_configuration"].pop("split_beam_ammo")

    specific_levels: Dict[str, str] = preset["layout_configuration"]["trick_level"]["specific_levels"]
    tricks_to_remove = [trick_name for trick_name, level in specific_levels.items() if level == "no-tricks"]
    for trick in tricks_to_remove:
        specific_levels.pop(trick)

    preset["game"] = preset["layout_configuration"].pop("game")
    preset["configuration"] = preset.pop("layout_configuration")
    preset["configuration"].update(preset.pop("patcher_configuration"))
    preset["configuration"]["varia_suit_damage"] = max(preset["configuration"]["varia_suit_damage"], 0.1)

    return preset


def _migrate_v6(preset: dict) -> dict:
    preset["configuration"]["dangerous_energy_tank"] = False
    return preset


def _migrate_v7(preset: dict) -> dict:
    default_items = {}
    if preset["game"] == "prime2":
        default_items["visor"] = "Combat Visor"
        default_items["beam"] = "Power Beam"

    preset["configuration"]["major_items_configuration"]["default_items"] = default_items
    return preset


_MIGRATIONS = {
    1: _migrate_v1,
    2: _migrate_v2,
    3: _migrate_v3,
    4: _migrate_v4,
    5: _migrate_v5,
    6: _migrate_v6,
    7: _migrate_v7,
}


def _apply_migration(preset: dict, version: int) -> dict:
    while version < CURRENT_PRESET_VERSION:
        preset = _MIGRATIONS[version](preset)
        version += 1
    return preset


def convert_to_current_version(preset: dict) -> dict:
    schema_version = preset["schema_version"]
    if schema_version > CURRENT_PRESET_VERSION:
        raise ValueError(f"Unknown version: {schema_version}")

    if schema_version < CURRENT_PRESET_VERSION:
        return _apply_migration(preset, schema_version)
    else:
        return preset


class VersionedPreset:
    data: dict
    exception: Optional[InvalidPreset] = None
    _preset: Optional[Preset] = None

    def __init__(self, data):
        self.data = data

    @classmethod
    def file_extension(cls) -> str:
        return "rdvpreset"

    @property
    def slug_name(self) -> str:
        return slugify.slugify(self.name)

    @property
    def name(self) -> str:
        if self.data is None:
            return self._preset.name
        else:
            return self.data["name"]

    @property
    def game(self) -> RandovaniaGame:
        if self.data is None:
            return self._preset.configuration.game

        if self.data["schema_version"] < 6:
            return RandovaniaGame.PRIME2

        return RandovaniaGame(self.data["game"])

    def __eq__(self, other):
        if isinstance(other, VersionedPreset):
            return self.get_preset() == other.get_preset()
        return False

    @property
    def _converted(self):
        return self._preset is not None or self.exception is not None

    def ensure_converted(self):
        if not self._converted:
            try:
                self._preset = Preset.from_json_dict(convert_to_current_version(copy.deepcopy(self.data)))
            except (ValueError, KeyError) as e:
                self.exception = InvalidPreset(e)
                raise self.exception from e

    def get_preset(self) -> Preset:
        self.ensure_converted()
        if self.exception:
            raise self.exception
        else:
            return self._preset

    @classmethod
    async def from_file(cls, path: Path) -> "VersionedPreset":
        async with aiofiles.open(path) as f:
            return VersionedPreset(json.loads(await f.read()))

    @classmethod
    def from_file_sync(cls, path: Path) -> "VersionedPreset":
        with path.open() as f:
            return VersionedPreset(json.load(f))

    @classmethod
    def with_preset(cls, preset: Preset) -> "VersionedPreset":
        result = VersionedPreset(None)
        result._preset = preset
        return result

    def save_to_file(self, path: Path):
        path.parent.mkdir(exist_ok=True, parents=True)
        with path.open("w") as preset_file:
            json.dump(self.as_json, preset_file, indent=4)

    @property
    def as_json(self) -> dict:
        if self._preset is not None:
            preset_json = {
                "schema_version": CURRENT_PRESET_VERSION,
            }
            preset_json.update(self._preset.as_json)
            return preset_json
        else:
            assert self.data is not None
            return self.data



import functools

from PySide2 import QtWidgets, QtCore
from PySide2.QtWidgets import QComboBox

from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.games.prime import default_data
from randovania.gui.generated.preset_echoes_translators_ui import Ui_PresetEchoesTranslators
from randovania.gui.lib.common_qt_lib import set_combo_with_value
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.enum_lib import iterate_enum
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset
from randovania.layout.translator_configuration import LayoutTranslatorRequirement


class PresetEchoesTranslators(PresetTab, Ui_PresetEchoesTranslators):
    _editor: PresetEditor

    def __init__(self, editor: PresetEditor):
        super().__init__()
        self.setupUi(self)
        self._editor = editor

        randomizer_data = default_data.decode_randomizer_data()

        self.translators_layout.setAlignment(QtCore.Qt.AlignTop)
        self.translator_randomize_all_button.clicked.connect(self._on_randomize_all_gates_pressed)
        self.translator_vanilla_actual_button.clicked.connect(self._on_vanilla_actual_gates_pressed)
        self.translator_vanilla_colors_button.clicked.connect(self._on_vanilla_colors_gates_pressed)

        self._combo_for_gate = {}

        for i, gate in enumerate(randomizer_data["TranslatorLocationData"]):
            label = QtWidgets.QLabel(self.translators_scroll_contents)
            label.setText(gate["Name"])
            self.translators_layout.addWidget(label, 3 + i, 0, 1, 1)

            combo = QComboBox(self.translators_scroll_contents)
            combo.gate = TranslatorGate(gate["Index"])
            for item in iterate_enum(LayoutTranslatorRequirement):
                combo.addItem(item.long_name, item)
            combo.currentIndexChanged.connect(functools.partial(self._on_gate_combo_box_changed, combo))

            self.translators_layout.addWidget(combo, 3 + i, 1, 1, 2)
            self._combo_for_gate[combo.gate] = combo

    @property
    def uses_patches_tab(self) -> bool:
        return True

    def _on_randomize_all_gates_pressed(self):
        with self._editor as editor:
            editor.set_configuration_field(
                "translator_configuration",
                editor.configuration.translator_configuration.with_full_random())

    def _on_vanilla_actual_gates_pressed(self):
        with self._editor as editor:
            editor.set_configuration_field(
                "translator_configuration",
                editor.configuration.translator_configuration.with_vanilla_actual())

    def _on_vanilla_colors_gates_pressed(self):
        with self._editor as editor:
            editor.set_configuration_field(
                "translator_configuration",
                editor.configuration.translator_configuration.with_vanilla_colors())

    def _on_gate_combo_box_changed(self, combo: QComboBox, new_index: int):
        with self._editor as editor:
            editor.set_configuration_field(
                "translator_configuration",
                editor.configuration.translator_configuration.replace_requirement_for_gate(
                    combo.gate, combo.currentData()))

    def on_preset_changed(self, preset: Preset):
        translator_configuration = preset.configuration.translator_configuration
        for gate, combo in self._combo_for_gate.items():
            set_combo_with_value(combo, translator_configuration.translator_requirement[gate])



import dataclasses
from enum import Enum
from typing import Optional

import pytest

from randovania.bitpacking.json_dataclass import JsonDataclass


class A(Enum):
    foo = "foo"
    bar = "bar"


@dataclasses.dataclass()
class D1(JsonDataclass):
    a: int
    b: str
    c: int = 5


@dataclasses.dataclass()
class D2(JsonDataclass):
    a: Optional[A]
    b: D1


@pytest.fixture(
    params=[
        {"instance": D2(a=A.bar, b=D1(a=5, b='foo', c=1)),
         "json": {'a': 'bar', 'b': {'a': 5, 'b': 'foo', 'c': 1}},
         },
        {"instance": D2(a=None, b=D1(a=5, b='foo', c=2)),
         "json": {'a': None, 'b': {'a': 5, 'b': 'foo', 'c': 2}},
         },
        {"instance": D2(a=None, b=D1(a=5, b='foo')),
         "json": {'a': None, 'b': {'a': 5, 'b': 'foo', 'c': 5}},
         }
    ],
    name="sample_values")
def _sample_values(request):
    yield request.param["instance"], request.param["json"]


def test_as_json(sample_values):
    value, data = sample_values
    assert value.as_json == data


def test_from_json(sample_values):
    value, data = sample_values
    assert D2.from_json(data) == value


def test_from_json_missing_field_with_default():
    value = D1(2, "foo")
    data = {"a": 2, "b": "foo"}
    assert D1.from_json(data) == value



import copy
from dataclasses import dataclass
from typing import Dict, Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder
from randovania.game_description.item.ammo import Ammo
from randovania.game_description.item.item_database import ItemDatabase
from randovania.layout.ammo_state import AmmoState


@dataclass(frozen=True)
class AmmoConfiguration(BitPackValue):
    maximum_ammo: Dict[int, int]
    items_state: Dict[Ammo, AmmoState]

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        default: AmmoConfiguration = metadata["reference"]

        assert list(self.maximum_ammo.keys()) == list(default.maximum_ammo.keys())
        assert list(self.items_state.keys()) == list(default.items_state.keys())

        for this, reference in zip(self.maximum_ammo.values(), default.maximum_ammo.values()):
            is_different = this != reference
            yield from bitpacking.encode_bool(is_different)
            if is_different:
                yield this, 256

        for this, reference in zip(self.items_state.values(), default.items_state.values()):
            is_different = this != reference
            yield from bitpacking.encode_bool(is_different)
            if is_different:
                yield from this.bit_pack_encode({})

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        default: AmmoConfiguration = metadata["reference"]

        # Maximum Ammo
        maximum_ammo = {}
        for item_key, default_value in default.maximum_ammo.items():
            is_different = bitpacking.decode_bool(decoder)
            if is_different:
                maximum_ammo[item_key] = decoder.decode_single(256)
            else:
                maximum_ammo[item_key] = default_value

        items_state = {}
        for item, default_state in default.items_state.items():
            is_different = bitpacking.decode_bool(decoder)
            if is_different:
                items_state[item] = AmmoState.bit_pack_unpack(decoder, {})
            else:
                items_state[item] = default_state

        return cls(maximum_ammo, items_state)

    @property
    def as_json(self) -> dict:
        return {
            "maximum_ammo": {
                str(ammo_item): maximum
                for ammo_item, maximum in self.maximum_ammo.items()
            },
            "items_state": {
                ammo.name: state.as_json
                for ammo, state in self.items_state.items()
            },
        }

    @classmethod
    def from_json(cls, value: dict, item_database: ItemDatabase) -> "AmmoConfiguration":
        return cls(
            maximum_ammo={
                int(ammo_item): maximum
                for ammo_item, maximum in value["maximum_ammo"].items()
            },
            items_state={
                item_database.ammo[name]: AmmoState.from_json(state)
                for name, state in value["items_state"].items()
            },
        )

    def replace_maximum_for_item(self, item: int, maximum: int) -> "AmmoConfiguration":
        return AmmoConfiguration(
            maximum_ammo={
                key: maximum if key == item else value
                for key, value in self.maximum_ammo.items()
            },
            items_state=copy.copy(self.items_state),
        )

    def replace_state_for_ammo(self, ammo: Ammo, state: AmmoState) -> "AmmoConfiguration":
        return self.replace_states({ammo: state})

    def replace_states(self, new_states: Dict[Ammo, AmmoState]) -> "AmmoConfiguration":
        """
        Creates a copy of this AmmoConfiguration where the state of all given items are replaced by the given
        states.
        :param new_states:
        :return:
        """
        items_state = copy.copy(self.items_state)

        for item, state in new_states.items():
            items_state[item] = state

        return AmmoConfiguration(copy.copy(self.maximum_ammo), items_state)



from pathlib import Path

import pytest
from mock import MagicMock, AsyncMock, call

import randovania
from randovania.network_client.network_client import NetworkClient, ConnectionState
from randovania.network_common.admin_actions import SessionAdminGlobalAction, SessionAdminUserAction
from randovania.network_common.error import InvalidSession


@pytest.fixture(name="client")
def _client(tmpdir):
    return NetworkClient(Path(tmpdir), {"server_address": "http://localhost:5000"})


@pytest.mark.asyncio
async def test_on_connect_no_restore(tmpdir):
    client = NetworkClient(Path(tmpdir), {"server_address": "http://localhost:5000"})

    # Run
    await client.on_connect()

    # Assert
    assert client.connection_state == ConnectionState.ConnectedNotLogged


@pytest.mark.parametrize("valid_session", [False, True])
@pytest.mark.asyncio
async def test_on_connect_restore(tmpdir, valid_session: bool):
    client = NetworkClient(Path(tmpdir), {"server_address": "http://localhost:5000"})
    session_data_path = Path(tmpdir) / "9iwAGnskOkqzo_NZ" / "session_persistence.bin"
    session_data_path.parent.mkdir(parents=True)
    session_data_path.write_bytes(b"foo")

    if valid_session:
        call_result = {"result": {
            "user": {
                "id": 1234,
                "name": "You",
            },
            "encoded_session_b85": b'Ze@30VtI6Ba{'
        }}
    else:
        call_result = InvalidSession().as_json

    client.sio = MagicMock()
    client.sio.call = AsyncMock(return_value=call_result)

    # Run
    await client.on_connect()

    # Assert
    client.sio.call.assert_awaited_once_with("restore_user_session", (b"foo", None), namespace=None, timeout=30)

    if valid_session:
        assert client.connection_state == ConnectionState.Connected
        assert session_data_path.read_bytes() == b"new_bytes"
    else:
        assert client.connection_state == ConnectionState.ConnectedNotLogged
        assert not session_data_path.is_file()


@pytest.mark.asyncio
async def test_connect_to_server(tmpdir):
    # Setup
    client = NetworkClient(Path(tmpdir), {"server_address": "http://localhost:5000",
                                          "socketio_path": "/path"})

    async def connect(*args, **kwargs):
        client._waiting_for_on_connect.set_result(True)

    client.sio = MagicMock()
    client.sio.connect = AsyncMock(side_effect=connect)
    client.sio.connected = False

    # Run
    await client.connect_to_server()

    # Assert
    assert client.connection_state == ConnectionState.Connecting
    client.sio.connect.assert_awaited_once_with("http://localhost:5000",
                                                socketio_path="/path",
                                                transports=["websocket"],
                                                headers={"X-Randovania-Version": randovania.VERSION})


@pytest.mark.asyncio
async def test_session_admin_global(client):
    client._emit_with_result = AsyncMock()
    client._current_game_session = MagicMock()
    client._current_game_session.id = 1234

    # Run
    result = await client.session_admin_global(SessionAdminGlobalAction.CHANGE_ROW, 5)

    # Assert
    assert result == client._emit_with_result.return_value
    client._emit_with_result.assert_awaited_once_with("game_session_admin_session", (1234, "change_row", 5))


@pytest.mark.parametrize("permanent", [False, True])
@pytest.mark.asyncio
async def test_leave_game_session(client: NetworkClient, permanent: bool):
    client._emit_with_result = AsyncMock()
    client._current_game_session = MagicMock()
    client._current_game_session.id = 1234
    client._current_user = MagicMock()
    client._current_user.id = 5678
    client._last_self_update = "foobar"

    # Run
    await client.leave_game_session(permanent)

    # Assert
    calls = [call("disconnect_game_session", 1234)]
    if permanent:
        calls.insert(0, call("game_session_admin_player", (1234, 5678, SessionAdminUserAction.KICK.value, None)))

    client._emit_with_result.assert_has_awaits(calls)

    assert client._current_game_session is None
    assert client._last_self_update is None



from pathlib import Path

import pytest
from PySide2 import QtWidgets
from mock import MagicMock, AsyncMock

from randovania.game_connection.dolphin_backend import DolphinBackend
from randovania.game_connection.game_connection import GameConnection
from randovania.gui.lib.game_connection_setup import GameConnectionSetup


@pytest.fixture(name="setup")
def _setup(skip_qtbot):
    parent = QtWidgets.QWidget()
    skip_qtbot.addWidget(parent)
    tool = QtWidgets.QToolButton(parent)
    label = QtWidgets.QLabel(parent)

    return GameConnectionSetup(parent, tool, label, GameConnection(DolphinBackend()), MagicMock())


@pytest.mark.parametrize("nintendont_ip", [None, "localhost", "192.168.0.1"])
@pytest.mark.asyncio
async def test_on_use_nintendont_backend_accept(setup, mocker, nintendont_ip):
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock,
                                       return_value=QtWidgets.QDialog.Accepted)
    setup.options.nintendont_ip = nintendont_ip
    old_backend = setup.game_connection.backend

    # Run
    await setup.on_use_nintendont_backend()

    # Assert
    mock_execute_dialog.assert_awaited_once()
    dialog: QtWidgets.QInputDialog = mock_execute_dialog.mock_calls[0].args[0]
    if nintendont_ip is not None:
        assert dialog.textValue() == nintendont_ip
        assert setup.game_connection.backend is not old_backend
        assert setup.use_nintendont_backend.isChecked()
        assert setup.use_nintendont_backend.text() == f"Nintendont: {nintendont_ip}"

    else:
        assert dialog.textValue() == ""
        assert setup.game_connection.backend is old_backend


@pytest.mark.asyncio
async def test_on_upload_nintendont_action_no_dol(setup, mocker, tmpdir):
    mocker.patch("randovania.gui.lib.game_connection_setup.get_data_path", return_value=Path(tmpdir))
    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)

    # Run
    await setup.on_upload_nintendont_action()

    # Assert
    execute_dialog.assert_awaited_once()


@pytest.mark.asyncio
async def test_on_upload_nintendont_action_with_dol(setup, mocker, tmpdir):
    mock_message_box = mocker.patch("PySide2.QtWidgets.QMessageBox")
    mocker.patch("randovania.gui.lib.game_connection_setup.get_data_path", return_value=Path(tmpdir))
    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)
    mock_upload = mocker.patch("wiiload.upload_file", new_callable=AsyncMock)

    nintendont_path = Path(tmpdir).joinpath("nintendont", "boot.dol")
    nintendont_path.parent.mkdir()
    nintendont_path.write_bytes(b"FOO")

    # Run
    await setup.on_upload_nintendont_action()

    # Assert
    execute_dialog.assert_not_awaited()
    mock_upload.assert_awaited_once_with(nintendont_path, [], setup.options.nintendont_ip)
    mock_message_box.assert_called_once()
    mock_message_box.return_value.show.assert_called_once_with()



import asyncio
import time
from argparse import ArgumentParser
from pathlib import Path

from randovania.cli import echoes_lib
from randovania.generator import generator
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.permalink import Permalink
from randovania.resolver import debug


async def _create_permalink(args) -> Permalink:
    from randovania.interface_common import persistence
    preset_manager = PresetManager(persistence.user_data_dir())
    await preset_manager.load_user_presets()
    preset = preset_manager.preset_for_name(args.preset_name).get_preset()

    return Permalink(
        args.seed_number,
        spoiler=True,
        presets={i: preset for i in range(args.player_count)},
    )


def distribute_command_logic(args):
    def status_update(s):
        if args.status_update:
            print(s)

    if args.permalink is not None:
        permalink = Permalink.from_str(args.permalink)
    else:
        permalink = asyncio.run(_create_permalink(args))
        print(f"Permalink: {permalink.as_base64_str}")

    if permalink.spoiler:
        debug.set_level(args.debug)

    extra_args = {}
    if args.no_retry:
        extra_args["attempts"] = 0

    before = time.perf_counter()
    layout_description = generator.generate_description(permalink=permalink, status_update=status_update,
                                                        validate_after_generation=args.validate, timeout=None,
                                                        **extra_args)
    after = time.perf_counter()
    print("Took {} seconds. Hash: {}".format(after - before, layout_description.shareable_hash))

    layout_description.save_to_file(args.output_file)


def add_distribute_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "distribute",
        help="Distribute pickups."
    )

    echoes_lib.add_debug_argument(parser)
    echoes_lib.add_validate_argument(parser)
    parser.add_argument("--no-retry", default=False, action="store_true", help="Disable retries in the generation.")
    parser.add_argument("--status-update", default=False, action="store_true", help="Print the status updates.")

    group = parser.add_mutually_exclusive_group()
    group.add_argument("--permalink", type=str, help="The permalink to use")
    group.add_argument("--preset-name", type=str, help="The name of the preset to use")

    parser.add_argument("--seed-number", type=int, default=0, help="If using a preset, the seed number. Defaults to 0.")
    parser.add_argument("--player-count", type=int, default=1,
                        help="If using a preset, the number of players. Defaults to 1.")

    parser.add_argument(
        "output_file",
        type=Path,
        help="Where to place the seed log.")
    parser.set_defaults(func=distribute_command_logic)



import hashlib
from typing import List

from randovania.game_description.area_location import AreaLocation
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_info import CurrentResources

LAYOUT_LETTERS = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ(){}[]<>=,.!#^-+?"
ITEM_NAME_TO_INDEX = {
    "Power Beam": 0,
    "Plasma Beam": 1,
    "Nova Beam": 2,
    "Charge Beam": 3,

    "Missile Launcher": 1000,
    "Ice Missile": 5,
    "Seeker Missile": 6,

    "Grapple Lasso": 7,
    "Grapple Swing": 8,
    "Grapple Voltage": 9,

    "Combat Visor": 11,
    "Scan Visor": 12,
    "Command Visor": 13,
    "X-Ray Visor": 14,

    "Space Jump Boots": 15,
    "Screw Attack": 16,

    "Hazard Shield": 17,
    "Energy Tank": 20,

    "Morph Ball": 32,
    "Morph Ball Bombs": 10,
    "Boost Ball": 33,
    "Spider Ball": 34,
    "CannonBall": 48,

    "Hypermode": 35,
    "Hyper Missile": 37,
    "Hyper Ball": 38,
    "Hyper Grapple": 39,

    "Ship Grapple": 44,
    "Ship Missile": 1001,
    "Missile Expansion": 4,
    "Ship Missile Expansion": 45
}
LETTER_TO_ITEM_MAPPING = {
    "0": ["Power Beam"],
    "1": ["Plasma Beam"],
    "2": ["Nova Beam"],
    "3": ["Charge Beam"],
    "4": ["Missile Expansion"],
    "5": ["Ice Missile"],
    "6": ["Seeker Missile"],
    "7": ["Grapple Lasso"],
    "8": ["Grapple Swing"],
    "9": ["Grapple Voltage"],
    "a": ["Morph Ball Bombs"],
    "b": ["Combat Visor"],
    "c": ["Scan Visor"],
    "d": ["Command Visor"],
    "e": ["X-Ray Visor"],
    "f": ["Space Jump Boots"],
    "g": ["Screw Attack"],
    "h": ["Hazard Shield"],
    # "i": ["Energy"],
    # "j": ["HyperModeEnergy"],
    "k": ["Energy Tank"],
    # "l": ["ItemPercentage"],
    # "m": ["Fuses"],
    "n": ["Energy Cell 1"],
    "o": ["Energy Cell 2"],
    "p": ["Energy Cell 3"],
    "q": ["Energy Cell 4"],
    "r": ["Energy Cell 5"],
    "s": ["Energy Cell 6"],
    "t": ["Energy Cell 7"],
    "u": ["Energy Cell 8"],
    "v": ["Energy Cell 9"],
    "w": ["Morph Ball"],
    "x": ["Boost Ball"],
    "y": ["Spider Ball"],
    "z": ["Hypermode"],
    # "A": ["HyperModeBeam"],
    "B": ["Hyper Missile"],
    "C": ["Hyper Ball"],
    "D": ["Hyper Grapple"],
    # "E": ["HyperModePermanent"],
    # "F": ["HyperModePhaaze"],
    # "G": ["HyperModeOriginal"],
    "H": ["Ship Grapple"],
    "I": ["Ship Missile Expansion"],
    # "J": ["FaceCorruptionLevel"],
    # "K": ["PhazonBall"],
    "L": ["CannonBall"],
    # "M": ["ActivateMorphballBoost"],
    # "N": ["HyperShot"],
    # "O": ["CommandVisorJammed"],
    # "P": ["Stat_Enemies_Killed"],
    # "Q": ["Stat_ShotsFired"],
    # "R": ["Stat_DamageReceived"],
    # "S": ["Stat_DataSaves"],
    # "T": ["Stat_HypermodeUses"],
    # "U": ["Stat_CommandoKills"],
    # "V": ["Stat_TinCanHighScore"],
    # "W": ["Stat_TinCanCurrentScore"],
    "X": ["Missile Expansion"] * 2,
    "Y": ["Missile Expansion"] * 3,
    "Z": ["Missile Expansion"] * 4,
    "(": ["Missile Expansion"] * 5,
    ")": ["Missile Expansion"] * 6,
    "{": ["Progressive Missile"],
    "}": ["Progressive Beam"],
    "[": ["Energy Tank", "Energy Tank"],
    "]": ["Energy Tank", "Energy Tank", "Energy Tank"],
    "<": ["Ship Missile Expansion", "Ship Missile Expansion"],
    ">": ["Missile Expansion", "Energy Tank"],
    "=": ["Energy Tank", "Missile Expansion"],
    ",": ["Missile Expansion", "Energy Tank", "Missile Expansion"],
    ".": ["Missile Expansion", "Ship Missile Expansion"],
    "!": ["Ship Missile Expansion", "Missile Expansion"],
    "#": ["Missile Launcher"],
    "^": ["Ship Missile"],
    "-": ["Ship Missile Expansion", "Energy Tank"],
    "+": ["Energy Tank", "Ship Missile Expansion"],
    "?": ["Missile Expansion", "Ship Missile Expansion", "Missile Expansion"],
}
ITEM_NAME_TO_LETTER_MAPPING = {
    items[0]: letter
    for letter, items in LETTER_TO_ITEM_MAPPING.items()
    if len(items) == 1
}

STARTING_ITEMS_ORDER = [
    "EnergyAmount",
    "EnergyCapacity",
    "EnergyTankAmount",
    "EnergyTankCapacity",
    "FusesAmount",
    "FusesCapacity",
    "PlayerInventoryItemAmount",
    "PlayerInventoryItemCapacity",
    "PowerBeam",
    "PlasmaBeam",
    "NovaBeam",
    "ChargeUpgrade",
    "MissileAmount",
    "MissileCapacity",
    "IceMissile",
    "Seekers",
    "GrappleBeamPull",
    "GrappleBeamSwing",
    "GrappleBeamVoltage",
    "Bomb",
    "ScanVisor",
    "CommandVisor",
    "XRayVisor",
    "MorphBall",
    "BoostBall",
    "SpiderBall",
    "DoubleJump",
    "SuitType",
    "ScrewAttack",
    "HyperModeTank",
    "HyperModeBeam",
    "HyperModeGrapple",
    "HyperModeMissile",
    "HyperModeBall",
    "HyperModePermanent",
    "HyperModePhaaze",
    "HyperModeOriginal",
    "HyperModeCharge",
    "ShipMissileAmount",
    "ShipMissileCapacity",
    "ShipGrapple",
]
STARTING_ITEMS_NAME_ALIAS = {
    "EnergyAmount": "Energy",
    "EnergyCapacity": "Energy",
    "EnergyTankAmount": "EnergyTank",
    "EnergyTankCapacity": "EnergyTank",
    "FusesAmount": "Fuses",
    "FusesCapacity": "Fuses",
    "MissileAmount": "Missile",
    "MissileCapacity": "Missile",
    "ShipMissileAmount": "ShipMissile",
    "ShipMissileCapacity": "ShipMissile",
}
_TWO_BYTE_VALUES = {STARTING_ITEMS_ORDER.index("MissileAmount"), STARTING_ITEMS_ORDER.index("MissileCapacity")}


def layout_string_for_items(item_names: List[str]) -> str:
    letters = [LAYOUT_LETTERS[0]] * 2
    for item in item_names:
        letters.append(ITEM_NAME_TO_LETTER_MAPPING[item])

    result = "".join(letters)
    sha = hashlib.sha256(result.encode("ascii"))
    result += sha.hexdigest()[:5]

    return result


def starting_location_for(location: AreaLocation) -> str:
    return f"custom {location.world_asset_id:x} {location.area_asset_id:x}"


def starting_items_for(resources: CurrentResources) -> str:
    capacity_by_short_name = {
        item.short_name: capacity
        for item, capacity in resources.items()
        if isinstance(item, ItemResourceInfo)
    }

    result_values = [
        capacity_by_short_name.get(STARTING_ITEMS_NAME_ALIAS.get(item, item), 0)
        for item in STARTING_ITEMS_ORDER
    ]
    return "custom " + "".join([
        "{:02x}".format(value) if index in _TWO_BYTE_VALUES else "{:x}".format(value)
        for index, value in enumerate(result_values)
    ])



import pytest

import randovania.generator.item_pool.ammo
import randovania.generator.item_pool.pickup_creator
from randovania.game_description.item.ammo import Ammo
from randovania.game_description.item.item_category import ItemCategory
from randovania.layout.ammo_state import AmmoState


@pytest.mark.parametrize(["per_pickup", "total_pickup", "included"], [
    (1, 5, 0),
    (1, 5, 1),
    (2, 5, 1),
    (50, 1, 0),
])
def test_items_for_ammo_one_item(per_pickup: int, total_pickup: int, included: int):
    # Setup
    item_a = 1

    included_ammo_for_item = {item_a: included}
    previous_pickup_for_item = {}

    maximum = per_pickup * total_pickup + included
    ammo = Ammo("My Ammo", maximum=maximum, items=(item_a,), broad_category=ItemCategory.BEAM_RELATED)
    state = AmmoState(0, total_pickup)
    maximum_ammo = {item_a: maximum}

    # Run
    ammo_per_pickup = randovania.generator.item_pool.ammo.items_for_ammo(ammo, state, included_ammo_for_item,
                                                                         previous_pickup_for_item, maximum_ammo)

    # Assert
    assert previous_pickup_for_item == {
        item_a: ammo
    }
    assert ammo_per_pickup == [[per_pickup]] * total_pickup


def test_items_for_ammo_one_item_non_divisible():
    # Setup
    item_a = 1

    maximum = 11
    total_pickup = 5

    included_ammo_for_item = {item_a: 0}
    previous_pickup_for_item = {}

    ammo = Ammo("My Ammo", maximum=maximum, items=(item_a,), broad_category=ItemCategory.BEAM_RELATED)
    state = AmmoState(0, total_pickup)
    maximum_ammo = {item_a: maximum}

    # Run
    ammo_per_pickup = randovania.generator.item_pool.ammo.items_for_ammo(ammo, state, included_ammo_for_item,
                                                                         previous_pickup_for_item, maximum_ammo)

    # Assert
    assert previous_pickup_for_item == {
        item_a: ammo
    }
    assert ammo_per_pickup == [[3]] + [[2]] * (total_pickup - 1)


@pytest.mark.parametrize(["per_pickup", "total_pickup", "included"], [
    (1, 5, 0),
    (1, 5, 1),
    (2, 5, 1),
    (50, 1, 0),
])
def test_items_for_ammo_two_item(per_pickup: int, total_pickup: int, included: int):
    # Setup
    item_a = 1
    item_b = 2

    included_ammo_for_item = {item_a: included, item_b: included}
    previous_pickup_for_item = {}

    maximum = per_pickup * total_pickup + included
    ammo = Ammo("My Ammo", maximum=maximum, items=(item_a, item_b), broad_category=ItemCategory.BEAM_RELATED)
    state = AmmoState(0, total_pickup)
    maximum_ammo = {item_a: maximum, item_b: maximum}

    # Run
    ammo_per_pickup = randovania.generator.item_pool.ammo.items_for_ammo(ammo, state, included_ammo_for_item,
                                                                         previous_pickup_for_item, maximum_ammo)

    # Assert
    assert previous_pickup_for_item == {
        item_a: ammo,
        item_b: ammo,
    }
    assert ammo_per_pickup == [[per_pickup, per_pickup]] * total_pickup


def test_items_for_ammo_two_item_diverging_values():
    # Setup
    item_a = 1
    item_b = 2

    total_pickup = 10
    maximum = 200

    included_ammo_for_item = {item_a: 0, item_b: 100}
    previous_pickup_for_item = {}

    ammo = Ammo("My Ammo", maximum=maximum, items=(item_a, item_b), broad_category=ItemCategory.BEAM_RELATED)
    state = AmmoState(0, total_pickup)
    maximum_ammo = {item_a: maximum, item_b: maximum}

    # Run
    ammo_per_pickup = randovania.generator.item_pool.ammo.items_for_ammo(ammo, state, included_ammo_for_item,
                                                                         previous_pickup_for_item, maximum_ammo)

    # Assert
    assert previous_pickup_for_item == {
        item_a: ammo,
        item_b: ammo,
    }
    assert ammo_per_pickup == [[20, 10]] * total_pickup



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'game_session.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_GameSessionWindow(object):
    def setupUi(self, GameSessionWindow):
        if not GameSessionWindow.objectName():
            GameSessionWindow.setObjectName(u"GameSessionWindow")
        GameSessionWindow.resize(699, 489)
        self.action_add_player = QAction(GameSessionWindow)
        self.action_add_player.setObjectName(u"action_add_player")
        self.action_add_row = QAction(GameSessionWindow)
        self.action_add_row.setObjectName(u"action_add_row")
        self.central_widget = QWidget(GameSessionWindow)
        self.central_widget.setObjectName(u"central_widget")
        self.central_widget.setMaximumSize(QSize(16777215, 16777215))
        self.main_layout = QVBoxLayout(self.central_widget)
        self.main_layout.setSpacing(6)
        self.main_layout.setContentsMargins(11, 11, 11, 11)
        self.main_layout.setObjectName(u"main_layout")
        self.title_layout = QHBoxLayout()
        self.title_layout.setSpacing(6)
        self.title_layout.setObjectName(u"title_layout")
        self.session_name_label = QLabel(self.central_widget)
        self.session_name_label.setObjectName(u"session_name_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.session_name_label.sizePolicy().hasHeightForWidth())
        self.session_name_label.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.session_name_label.setFont(font)

        self.title_layout.addWidget(self.session_name_label)

        self.advanced_options_tool = QToolButton(self.central_widget)
        self.advanced_options_tool.setObjectName(u"advanced_options_tool")
        self.advanced_options_tool.setPopupMode(QToolButton.InstantPopup)

        self.title_layout.addWidget(self.advanced_options_tool)


        self.main_layout.addLayout(self.title_layout)

        self.players_box = QGroupBox(self.central_widget)
        self.players_box.setObjectName(u"players_box")
        sizePolicy1 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.players_box.sizePolicy().hasHeightForWidth())
        self.players_box.setSizePolicy(sizePolicy1)
        self.players_layout = QGridLayout(self.players_box)
        self.players_layout.setSpacing(6)
        self.players_layout.setContentsMargins(11, 11, 11, 11)
        self.players_layout.setObjectName(u"players_layout")
        self.new_row_button = QPushButton(self.players_box)
        self.new_row_button.setObjectName(u"new_row_button")

        self.players_layout.addWidget(self.new_row_button, 0, 0, 1, 2)

        self.title_connection_state_label = QLabel(self.players_box)
        self.title_connection_state_label.setObjectName(u"title_connection_state_label")

        self.players_layout.addWidget(self.title_connection_state_label, 0, 5, 1, 1)

        self.players_vertical_line = QFrame(self.players_box)
        self.players_vertical_line.setObjectName(u"players_vertical_line")
        self.players_vertical_line.setFrameShape(QFrame.VLine)
        self.players_vertical_line.setFrameShadow(QFrame.Sunken)

        self.players_layout.addWidget(self.players_vertical_line, 0, 2, 2, 1)

        self.title_player_name_label = QLabel(self.players_box)
        self.title_player_name_label.setObjectName(u"title_player_name_label")

        self.players_layout.addWidget(self.title_player_name_label, 0, 4, 1, 1)

        self.presets_line = QFrame(self.players_box)
        self.presets_line.setObjectName(u"presets_line")
        self.presets_line.setFrameShape(QFrame.HLine)
        self.presets_line.setFrameShadow(QFrame.Sunken)

        self.players_layout.addWidget(self.presets_line, 1, 0, 1, 2)

        self.team_line = QFrame(self.players_box)
        self.team_line.setObjectName(u"team_line")
        self.team_line.setFrameShape(QFrame.HLine)
        self.team_line.setFrameShadow(QFrame.Sunken)

        self.players_layout.addWidget(self.team_line, 1, 4, 1, 3)


        self.main_layout.addWidget(self.players_box)

        self.observer_group = QGroupBox(self.central_widget)
        self.observer_group.setObjectName(u"observer_group")
        sizePolicy1.setHeightForWidth(self.observer_group.sizePolicy().hasHeightForWidth())
        self.observer_group.setSizePolicy(sizePolicy1)
        self.observer_layout = QGridLayout(self.observer_group)
        self.observer_layout.setSpacing(6)
        self.observer_layout.setContentsMargins(11, 11, 11, 11)
        self.observer_layout.setObjectName(u"observer_layout")

        self.main_layout.addWidget(self.observer_group)

        self.tab_widget = QTabWidget(self.central_widget)
        self.tab_widget.setObjectName(u"tab_widget")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.tab_widget.sizePolicy().hasHeightForWidth())
        self.tab_widget.setSizePolicy(sizePolicy2)
        self.tab_generate = QWidget()
        self.tab_generate.setObjectName(u"tab_generate")
        sizePolicy3 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.tab_generate.sizePolicy().hasHeightForWidth())
        self.tab_generate.setSizePolicy(sizePolicy3)
        self.gridLayout = QGridLayout(self.tab_generate)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.view_game_details_button = QPushButton(self.tab_generate)
        self.view_game_details_button.setObjectName(u"view_game_details_button")

        self.gridLayout.addWidget(self.view_game_details_button, 1, 4, 1, 1)

        self.generate_game_label = QLabel(self.tab_generate)
        self.generate_game_label.setObjectName(u"generate_game_label")

        self.gridLayout.addWidget(self.generate_game_label, 1, 0, 1, 3)

        self.progress_label = QLabel(self.tab_generate)
        self.progress_label.setObjectName(u"progress_label")
        sizePolicy4 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.progress_label.sizePolicy().hasHeightForWidth())
        self.progress_label.setSizePolicy(sizePolicy4)
        self.progress_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)
        self.progress_label.setWordWrap(True)

        self.gridLayout.addWidget(self.progress_label, 4, 1, 1, 4)

        self.background_process_button = QToolButton(self.tab_generate)
        self.background_process_button.setObjectName(u"background_process_button")
        self.background_process_button.setMinimumSize(QSize(140, 0))
        self.background_process_button.setPopupMode(QToolButton.MenuButtonPopup)

        self.gridLayout.addWidget(self.background_process_button, 4, 0, 1, 1)

        self.customize_user_preferences_button = QPushButton(self.tab_generate)
        self.customize_user_preferences_button.setObjectName(u"customize_user_preferences_button")

        self.gridLayout.addWidget(self.customize_user_preferences_button, 1, 6, 1, 1)

        self.line_generate = QFrame(self.tab_generate)
        self.line_generate.setObjectName(u"line_generate")
        self.line_generate.setFrameShape(QFrame.HLine)
        self.line_generate.setFrameShadow(QFrame.Sunken)

        self.gridLayout.addWidget(self.line_generate, 3, 0, 1, 7)

        self.progress_bar = QProgressBar(self.tab_generate)
        self.progress_bar.setObjectName(u"progress_bar")
        self.progress_bar.setValue(0)
        self.progress_bar.setInvertedAppearance(False)

        self.gridLayout.addWidget(self.progress_bar, 4, 5, 1, 2)

        self.save_iso_button = QToolButton(self.tab_generate)
        self.save_iso_button.setObjectName(u"save_iso_button")
        self.save_iso_button.setMinimumSize(QSize(100, 0))
        self.save_iso_button.setPopupMode(QToolButton.MenuButtonPopup)

        self.gridLayout.addWidget(self.save_iso_button, 1, 5, 1, 1)

        self.tab_widget.addTab(self.tab_generate, "")
        self.tab_history = QWidget()
        self.tab_history.setObjectName(u"tab_history")
        sizePolicy3.setHeightForWidth(self.tab_history.sizePolicy().hasHeightForWidth())
        self.tab_history.setSizePolicy(sizePolicy3)
        self.history_layout = QVBoxLayout(self.tab_history)
        self.history_layout.setSpacing(6)
        self.history_layout.setContentsMargins(11, 11, 11, 11)
        self.history_layout.setObjectName(u"history_layout")
        self.history_layout.setSizeConstraint(QLayout.SetMaximumSize)
        self.history_layout.setContentsMargins(4, 4, 4, 4)
        self.history_table_widget = QTableWidget(self.tab_history)
        if (self.history_table_widget.columnCount() < 2):
            self.history_table_widget.setColumnCount(2)
        __qtablewidgetitem = QTableWidgetItem()
        self.history_table_widget.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.history_table_widget.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        self.history_table_widget.setObjectName(u"history_table_widget")
        self.history_table_widget.horizontalHeader().setVisible(False)
        self.history_table_widget.horizontalHeader().setDefaultSectionSize(200)
        self.history_table_widget.verticalHeader().setVisible(False)

        self.history_layout.addWidget(self.history_table_widget)

        self.tab_widget.addTab(self.tab_history, "")

        self.main_layout.addWidget(self.tab_widget)

        self.connection_group = QGroupBox(self.central_widget)
        self.connection_group.setObjectName(u"connection_group")
        self.connection_group.setFlat(False)
        self.connection_layout = QHBoxLayout(self.connection_group)
        self.connection_layout.setSpacing(6)
        self.connection_layout.setContentsMargins(11, 11, 11, 11)
        self.connection_layout.setObjectName(u"connection_layout")
        self.connection_layout.setContentsMargins(-1, 4, -1, 4)
        self.game_connection_tool = QToolButton(self.connection_group)
        self.game_connection_tool.setObjectName(u"game_connection_tool")
        self.game_connection_tool.setPopupMode(QToolButton.InstantPopup)

        self.connection_layout.addWidget(self.game_connection_tool)

        self.game_connection_label = QLabel(self.connection_group)
        self.game_connection_label.setObjectName(u"game_connection_label")

        self.connection_layout.addWidget(self.game_connection_label)

        self.connection_line_1 = QFrame(self.connection_group)
        self.connection_line_1.setObjectName(u"connection_line_1")
        self.connection_line_1.setFrameShape(QFrame.VLine)
        self.connection_line_1.setFrameShadow(QFrame.Sunken)

        self.connection_layout.addWidget(self.connection_line_1)

        self.server_connection_button = QPushButton(self.connection_group)
        self.server_connection_button.setObjectName(u"server_connection_button")
        sizePolicy5 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy5.setHorizontalStretch(0)
        sizePolicy5.setVerticalStretch(0)
        sizePolicy5.setHeightForWidth(self.server_connection_button.sizePolicy().hasHeightForWidth())
        self.server_connection_button.setSizePolicy(sizePolicy5)
        self.server_connection_button.setMaximumSize(QSize(60, 16777215))

        self.connection_layout.addWidget(self.server_connection_button)

        self.server_connection_label = QLabel(self.connection_group)
        self.server_connection_label.setObjectName(u"server_connection_label")

        self.connection_layout.addWidget(self.server_connection_label)

        self.connection_line_2 = QFrame(self.connection_group)
        self.connection_line_2.setObjectName(u"connection_line_2")
        self.connection_line_2.setFrameShape(QFrame.VLine)
        self.connection_line_2.setFrameShadow(QFrame.Sunken)

        self.connection_layout.addWidget(self.connection_line_2)

        self.session_status_tool = QToolButton(self.connection_group)
        self.session_status_tool.setObjectName(u"session_status_tool")
        self.session_status_tool.setMaximumSize(QSize(80, 16777215))
        self.session_status_tool.setPopupMode(QToolButton.MenuButtonPopup)

        self.connection_layout.addWidget(self.session_status_tool)

        self.session_status_label = QLabel(self.connection_group)
        self.session_status_label.setObjectName(u"session_status_label")

        self.connection_layout.addWidget(self.session_status_label)


        self.main_layout.addWidget(self.connection_group)

        GameSessionWindow.setCentralWidget(self.central_widget)
        self.menuBar = QMenuBar(GameSessionWindow)
        self.menuBar.setObjectName(u"menuBar")
        self.menuBar.setGeometry(QRect(0, 0, 699, 21))
        GameSessionWindow.setMenuBar(self.menuBar)

        self.retranslateUi(GameSessionWindow)

        self.tab_widget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(GameSessionWindow)
    # setupUi

    def retranslateUi(self, GameSessionWindow):
        GameSessionWindow.setWindowTitle(QCoreApplication.translate("GameSessionWindow", u"Game Session", None))
        self.action_add_player.setText(QCoreApplication.translate("GameSessionWindow", u"Add player", None))
        self.action_add_row.setText(QCoreApplication.translate("GameSessionWindow", u"Add row", None))
        self.session_name_label.setText(QCoreApplication.translate("GameSessionWindow", u"Session name!", None))
        self.advanced_options_tool.setText(QCoreApplication.translate("GameSessionWindow", u"Advanced options...", None))
        self.players_box.setTitle("")
        self.new_row_button.setText(QCoreApplication.translate("GameSessionWindow", u"New Row", None))
        self.title_connection_state_label.setText(QCoreApplication.translate("GameSessionWindow", u"Connection state", None))
        self.title_player_name_label.setText(QCoreApplication.translate("GameSessionWindow", u"Players", None))
#if QT_CONFIG(tooltip)
        self.observer_group.setToolTip(QCoreApplication.translate("GameSessionWindow", u"<html><head/><body><p>These users won't participate on the match, but they can be admins and follow trackers.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.observer_group.setTitle(QCoreApplication.translate("GameSessionWindow", u"Observers", None))
        self.view_game_details_button.setText(QCoreApplication.translate("GameSessionWindow", u"View Spoiler", None))
        self.generate_game_label.setText(QCoreApplication.translate("GameSessionWindow", u"<Game not generated>", None))
        self.progress_label.setText("")
        self.background_process_button.setText(QCoreApplication.translate("GameSessionWindow", u"Stop", None))
        self.customize_user_preferences_button.setText(QCoreApplication.translate("GameSessionWindow", u"Customize in-game settings", None))
        self.save_iso_button.setText(QCoreApplication.translate("GameSessionWindow", u"Save ISO", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_generate), QCoreApplication.translate("GameSessionWindow", u"Game", None))
        ___qtablewidgetitem = self.history_table_widget.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("GameSessionWindow", u"Message", None));
        ___qtablewidgetitem1 = self.history_table_widget.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("GameSessionWindow", u"Time", None));
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_history), QCoreApplication.translate("GameSessionWindow", u"History", None))
        self.connection_group.setTitle("")
        self.game_connection_tool.setText(QCoreApplication.translate("GameSessionWindow", u"Connect to game", None))
        self.game_connection_label.setText(QCoreApplication.translate("GameSessionWindow", u"Game: Disconnected", None))
        self.server_connection_button.setText(QCoreApplication.translate("GameSessionWindow", u"Connect", None))
        self.server_connection_label.setText(QCoreApplication.translate("GameSessionWindow", u"Server: Disconnected", None))
        self.session_status_tool.setText(QCoreApplication.translate("GameSessionWindow", u"Start", None))
        self.session_status_label.setText(QCoreApplication.translate("GameSessionWindow", u"Session: Not Started", None))
    # retranslateUi




import copy
from collections import defaultdict
from random import Random
from typing import List, Dict, Optional, Tuple, AbstractSet

from randovania.game_description.area_location import AreaLocation
from randovania.game_description.node import TeleporterNode
from randovania.game_description.world_list import WorldList


class Elevator:
    instance_id: int
    world_asset_id: int
    area_asset_id: int
    destination_world: int
    destination_area: int
    connected_elevator: Optional["Elevator"]

    def __init__(self, instance_id: int, world_asset_id: int, area_asset_id: int, destination_world: int,
                 destination_area: int):
        self.world_asset_id = world_asset_id
        self.area_asset_id = area_asset_id
        self.instance_id = instance_id
        self.destination_world = destination_world
        self.destination_area = destination_area
        self.connected_elevator = None

    def connect_to(self, other: "Elevator"):
        self.destination_world = other.world_asset_id
        self.destination_area = other.area_asset_id
        other.destination_world = self.world_asset_id
        other.destination_area = self.area_asset_id
        self.connected_elevator = other
        other.connected_elevator = self

    @property
    def area_location(self):
        return AreaLocation(self.world_asset_id, self.area_asset_id)


def try_randomize_elevators(rng: Random,
                            echoes_elevators: Tuple[Elevator, ...],
                            ) -> List[Elevator]:
    elevator_database: List[Elevator] = list(echoes_elevators)

    elevator_list = copy.copy(elevator_database)
    elevators_by_world: Dict[int, List[Elevator]] = defaultdict(list)
    for elevator in elevator_list:
        elevators_by_world[elevator.world_asset_id].append(elevator)

    while elevator_list:
        source_elevators: List[Elevator] = max(elevators_by_world.values(), key=len)
        target_elevators: List[Elevator] = [
            elevator
            for elevator in elevator_list
            if elevator not in source_elevators
        ]
        source_elevator = source_elevators[0]
        target_elevator = rng.choice(target_elevators)

        source_elevator.connect_to(target_elevator)

        elevators_by_world[source_elevator.world_asset_id].remove(source_elevator)
        elevators_by_world[target_elevator.world_asset_id].remove(target_elevator)
        elevator_list.remove(source_elevator)
        elevator_list.remove(target_elevator)

    # TODO
    list3 = copy.copy(elevator_database)
    celevator_list3 = [list3[0]]
    while list3:
        celevator_list1 = []
        for celevator1 in celevator_list3:
            index = 0
            while index < len(list3):
                celevator2 = list3[index]
                if celevator2.world_asset_id == celevator1.world_asset_id or celevator2.area_asset_id == celevator1.destination_area:
                    celevator_list1.append(celevator2)
                    list3.remove(celevator2)
                else:
                    index += 1
        if celevator_list1:
            celevator_list3 = celevator_list1
        else:
            # Randomization failed
            return try_randomize_elevators(rng, echoes_elevators)

    return elevator_database


def two_way_elevator_connections(rng: Random,
                                 elevator_database: Tuple[Elevator, ...],
                                 between_areas: bool
                                 ) -> Dict[int, AreaLocation]:
    if between_areas:
        elevator_database = try_randomize_elevators(rng, elevator_database)
    else:
        assert len(elevator_database) % 2 == 0
        elevators = list(elevator_database)
        rng.shuffle(elevators)
        while elevators:
            elevators.pop().connect_to(elevators.pop())

    return {
        elevator.instance_id: elevator.connected_elevator.area_location
        for elevator in elevator_database
    }


def one_way_elevator_connections(rng: Random,
                                 elevator_database: Tuple[Elevator, ...],
                                 world_list: WorldList,
                                 elevator_target: bool
                                 ) -> Dict[int, AreaLocation]:
    if elevator_target:
        target_locations = [elevator.area_location for elevator in elevator_database]
    else:
        target_locations = [
            AreaLocation(world.world_asset_id, area.area_asset_id)
            for world in world_list.worlds
            for area in world.areas
        ]

    rng.shuffle(target_locations)

    return {
        elevator.instance_id: target_locations.pop()
        for elevator in elevator_database
    }


def create_elevator_database(world_list: WorldList,
                             areas_to_not_change: AbstractSet[int],
                             ) -> Tuple[Elevator, ...]:
    """
    Creates a tuple of Elevator objects, exclude those that belongs to one of the areas provided.
    :param world_list:
    :param areas_to_not_change: Set of asset_id of Areas whose teleporters are to be ignored
    :return:
    """
    return tuple(
        Elevator(node.teleporter_instance_id,
                 world.world_asset_id,
                 area.area_asset_id,
                 node.default_connection.world_asset_id,
                 node.default_connection.area_asset_id)

        for world, area, node in world_list.all_worlds_areas_nodes
        if isinstance(node, TeleporterNode) and node.editable and area.area_asset_id not in areas_to_not_change
    )



from unittest.mock import patch, MagicMock

from randovania.cli.commands import batch_distribute
from randovania.layout.permalink import Permalink


@patch("randovania.generator.generator.generate_description", autospec=True)
@patch("time.perf_counter", autospec=False)  # TODO: pytest-qt bug
def test_batch_distribute_helper(mock_perf_counter: MagicMock,
                                 mock_generate_description: MagicMock,
                                 ):
    # Setup
    base_permalink = MagicMock()
    seed_number = 5000
    validate = MagicMock()
    output_dir = MagicMock()
    timeout = 67

    expected_permalink = Permalink(
        seed_number=seed_number,
        spoiler=True,
        presets=base_permalink.presets,
    )

    mock_perf_counter.side_effect = [1000, 5000]

    # Run
    delta_time = batch_distribute.batch_distribute_helper(base_permalink, seed_number, timeout, validate, output_dir)

    # Assert
    mock_generate_description.assert_called_once_with(permalink=expected_permalink, status_update=None,
                                                      validate_after_generation=validate, timeout=timeout,
                                                      attempts=0)
    assert delta_time == 4000
    output_dir.joinpath.assert_called_once_with("{}.json".format(seed_number))
    mock_generate_description.return_value.save_to_file.assert_called_once_with(output_dir.joinpath.return_value)



from randovania.games.game import RandovaniaGame
from randovania.games.prime.all_prime_dol_patches import StringDisplayPatchAddresses, HealthCapacityAddresses, \
    DangerousEnergyTankAddresses, PowerupFunctionsAddresses
from randovania.games.prime.echoes_dol_patches import BeamCostAddresses, SafeZoneAddresses, EchoesDolVersion, \
    StartingBeamVisorAddresses

ALL_VERSIONS = [
    EchoesDolVersion(
        game=RandovaniaGame.PRIME2,
        description="Gamecube NTSC",
        build_string_address=0x803ac3b0,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32",
        sda2_base=0x804223c0,
        game_state_pointer=0x80418eb8,
        cplayer_vtable=0x803b15d0,
        cstate_manager_global=0x803db6e0,
        string_display=StringDisplayPatchAddresses(
            update_hint_state=0x80038020,
            message_receiver_string_ref=0x803bd118,
            wstring_constructor=0x802ff3dc,
            display_hud_memo=0x8006b3c8,
            max_message_size=200,
        ),
        powerup_functions=PowerupFunctionsAddresses(
            add_power_up=0x800858f0,
            incr_pickup=0x80085760,
            decr_pickup=0x800856c4,
        ),
        health_capacity=HealthCapacityAddresses(
            base_health_capacity=0x8041abe4,
            energy_tank_capacity=0x8041abe0,
        ),
        dangerous_energy_tank=DangerousEnergyTankAddresses(
            small_number_float=0x8041a4a8,
            incr_pickup=0x80085760,
        ),
        beam_cost_addresses=BeamCostAddresses(
            uncharged_cost=0x803aa8c8,
            charged_cost=0x803aa8d8,
            charge_combo_ammo_cost=0x803aa8e8,
            charge_combo_missile_cost=0x803a74ac,
            get_beam_ammo_type_and_costs=0x801cccb0,
        ),
        game_options_constructor_address=0x80161b48,
        safe_zone=SafeZoneAddresses(
            heal_per_frame_constant=0x8041a4fc,
            increment_health_fmr=0x8000c710,
        ),
        starting_beam_visor=StartingBeamVisorAddresses(
            player_state_constructor_clean=0x80086008,
            player_state_constructor_decode=0x80085c20,
            health_info_constructor=0x801420c8,
            enter_morph_ball_state=0x80184118,
            start_transition_to_visor=0x80085314,
        )
    ),
    EchoesDolVersion(
        game=RandovaniaGame.PRIME2,
        description="Gamecube PAL",
        build_string_address=0x803ad710,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.035 10/27/2004 19:48:17",
        sda2_base=0x804223c0,
        game_state_pointer=0x8041A19C,
        cplayer_vtable=0x803b2950,
        cstate_manager_global=0x803dc900,
        string_display=StringDisplayPatchAddresses(
            update_hint_state=0x80038194,
            message_receiver_string_ref=0x803be378,
            wstring_constructor=0x802ff734,
            display_hud_memo=0x8006b504,
            max_message_size=200,
        ),
        powerup_functions=PowerupFunctionsAddresses(
            add_power_up=0x80085a2c,
            incr_pickup=0x8008589c,
            decr_pickup=0x80085800,
        ),
        health_capacity=HealthCapacityAddresses(
            base_health_capacity=0x8041bedc,
            energy_tank_capacity=0x8041bed8,
        ),
        dangerous_energy_tank=DangerousEnergyTankAddresses(
            small_number_float=0x8041b7a0,
            incr_pickup=0x8008589c,
        ),
        beam_cost_addresses=BeamCostAddresses(
            uncharged_cost=0x803abc28,
            charged_cost=0x803abc38,
            charge_combo_ammo_cost=0x803abc48,
            charge_combo_missile_cost=0x803a7c04,
            get_beam_ammo_type_and_costs=0x801ccfe4,
        ),
        game_options_constructor_address=0x80161d9c,
        safe_zone=SafeZoneAddresses(
            heal_per_frame_constant=0x8041b7f4,
            increment_health_fmr=0x8000c754,
        ),
        starting_beam_visor=StartingBeamVisorAddresses(
            player_state_constructor_clean=0x80086144,
            player_state_constructor_decode=0x80085d5c,
            health_info_constructor=0x80142304,
            enter_morph_ball_state=0x801843f8,
            start_transition_to_visor=0x80085450,
        )
    ),
]



from randovania.gui.corruption_layout_editor import CorruptionLayoutEditor


def test_construct(skip_qtbot):
    expected = ("00xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
                "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx3b95e")
    editor = CorruptionLayoutEditor()
    assert editor.layout_edit.text() == expected



import argparse
import asyncio
import logging.config
import os
import sys

from PySide2.QtCore import QCoreApplication

from randovania.game_connection.connection_backend import ConnectionBackend
from randovania.game_connection.connection_base import GameConnectionStatus
from randovania.game_connection.dolphin_backend import DolphinBackend
from randovania.game_connection.game_connection import GameConnection
from randovania.game_connection.nintendont_backend import NintendontBackend

should_quit = False
old_hook = sys.excepthook


async def worker(app: QCoreApplication, backend: ConnectionBackend):
    connection = GameConnection(backend)

    await connection.start()
    try:
        while connection.current_status not in (GameConnectionStatus.InGame, GameConnectionStatus.TrackerOnly):
            await asyncio.sleep(1)
            print(connection.pretty_current_status)

        await asyncio.sleep(20)

    except KeyboardInterrupt:
        return

    finally:
        await connection.stop()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--nintendont", help="Connect to the given IP via the Nintendont protocol instead.")
    args = parser.parse_args()

    app = QCoreApplication(sys.argv)

    os.environ['QT_API'] = "PySide2"
    import asyncqt
    loop = asyncqt.QEventLoop(app)
    asyncio.set_event_loop(loop)

    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] [%(levelname)s] [%(name)s] %(funcName)s: %(message)s',
            }
        },
        'handlers': {
            'default': {
                'level': 'DEBUG',
                'formatter': 'default',
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stdout',  # Default is stderr
            },
        },
        'loggers': {
        },
        'root': {
            'level': 'DEBUG',
            'handlers': ['default'],
        },
    })

    def catch_exceptions(t, val, tb):
        global should_quit
        should_quit = True
        old_hook(t, val, tb)

    def catch_exceptions_async(loop, context):
        if 'future' in context:
            future: asyncio.Future = context['future']
            logging.exception(context["message"], exc_info=future.exception())
        else:
            logging.critical(str(context))

    sys.excepthook = catch_exceptions
    loop.set_exception_handler(catch_exceptions_async)

    if args.nintendont is not None:
        backend = NintendontBackend(args.nintendont)
    else:
        backend = DolphinBackend()

    with loop:
        try:
            asyncio.get_event_loop().run_until_complete(worker(app, backend))
        finally:
            app.quit()


if __name__ == '__main__':
    main()
